import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory
from typing import List, Optional, Union

from mlpype.base.constants import Constants
from mlpype.base.utils.workspace import switch_directory

from .extensions import WheelExtension

BaseExtension = WheelExtension(
    "base",
    functionality=[(Path(__file__).parent / "helpers" / "wheel_main.py", ["load_model"])],
    libraries=["mlpype.base"],
)


@dataclass
class WheelBuilder:
    """Class to help turn an output folder from an Experiment into a wheel file.

    This should allow you to save models into shippable packages
    quite easily.

    The functionality can be extended quite easily using the `extensions` argument.
    See `WheelExtension` for how to make your own. The base functionality is also
    included as an extension, feel free to check out how that was done as well.

    Args:
        model_folder (Path): Path to the output folder generated by an Experiment
        model_name (str): The name of the model. Will be used as package name. Should follow
            proper package naming conventions.
        version (str): Model/package version.
        output_wheel_file (Optional[Union[Path, str]]): The output directory. Defaults to "wheel_output"
            in the current directory.
        libraries (Optional[List[str]]): A list of (versioned) libraries that are used as
            dependencies. By default we use the requirements.txt file from the experiment run.
        extensions (Optional[List[WheelExtension]], optional): A list of extension to add to this
            wheel file. Extensions add extra functionality besides the base "import pipeline" functionality.
            This can include hosting a FastAPI app for instance.
    """

    # TODO: add proper tests?
    model_folder: Path
    model_name: str
    version: str
    output_wheel_file: Optional[Union[Path, str]] = None
    libraries: Optional[List[str]] = None
    extensions: Optional[List[WheelExtension]] = None

    def __post_init__(self) -> None:
        """Performs checks and default initialisations."""
        if self.output_wheel_file is None:
            self.output_wheel_file = Path("").parent.absolute() / "wheel_output"
        if self.extensions is None:
            self.extensions = []
        if BaseExtension not in self.extensions:
            self.extensions.append(BaseExtension)
        self._validate_extensions()

    def _validate_extensions(self) -> None:
        names = set()
        function_names = set()

        assert isinstance(self.extensions, list)
        for extension in self.extensions:
            assert extension.name not in names, f"{extension.name} is a duplicated extension name, this is not allowed."
            names.add(extension.name)

            for _, imports in extension.functionality:
                for imp in imports:
                    assert (
                        imp not in function_names
                    ), f"{imp} is a duplicated function name from {extension.name}, this is not allowed."
                    function_names.add(imp)

    def build(self) -> None:
        """Builds the actual package."""
        local_dir = Path(__file__).parent / "helpers"
        template_file = local_dir / "setup_template.py"

        if self.libraries is None:
            with open(self.model_folder / Constants.REQUIREMENTS_FILE, "r") as f:
                requirements_raw = f.readlines()
                libraries = list(map(lambda x: x.replace("\n", ""), filter(lambda x: "==" in x, requirements_raw)))
        else:
            libraries = self.libraries
        assert len(libraries) > 0, "Using mlpype, it is assumend there is at least 1 dependency"

        with TemporaryDirectory() as tmp_directory:
            tmp_directory_path = Path(tmp_directory)
            tmp_package_dir = tmp_directory_path / self.model_name
            tmp_model_dir = tmp_package_dir / "outputs"
            tmp_setup_file = tmp_directory_path / "setup.py"
            tmp_init_file = tmp_package_dir / "__init__.py"

            tmp_package_dir.mkdir()
            with open(tmp_init_file, "w") as f:
                f.write("")

            assert isinstance(self.extensions, list)
            for extension in self.extensions:
                print(f"Adding extension `{extension.name}`:", extension)
                extension.extend(self.model_name, libraries, tmp_package_dir)

            requirements = '", "'.join(libraries)

            with open(template_file, "r") as f:
                setup_file_raw = f.read()
                setup_file_formatted = (
                    setup_file_raw.replace("{install_requires}", requirements)
                    .replace("{package_name}", self.model_name)
                    .replace("{version}", self.version)
                )

            copytree(self.model_folder, tmp_model_dir)
            with open(tmp_setup_file, "w") as f:
                f.write(setup_file_formatted)

            output_wheel_file = str(self.output_wheel_file)
            with switch_directory(tmp_directory):
                subprocess.check_output(
                    [sys.executable, "setup.py", "bdist_wheel", "--dist-dir", output_wheel_file]
                ).decode()



from typing import Any

import pandas as pd

from mlpype.base.data.data_source import DataSource


class SqlSource(DataSource[pd.DataFrame]):
    def __init__(
        self,
        sql: str,
        con: str,
        *args: Any,
        **kwargs: Any,
    ):
        """Reads data from a sql table using Pandas read_sql.

        This class only works as a storage for the arguments to call Pandas read_sql.
        For proper documentation on the arguments, check Pandas read_sql.

        Args:
            sql (str): The sql and the `sql` argument to `Pandas read_sql`.
            con (str): The con string and `con` argument to `Pandas read_sql`.
            args: Positional arguments to be passed to `Pandas read_sql`.
            kwargs: Keyword arguments to be passed to `Pandas read_sql`.
        """
        self.sql = sql
        self.con = con
        self.args = args
        self.kwargs = kwargs
        super().__init__()

    def read(self) -> pd.DataFrame:
        """Use the provided arguments to call `read_sql`.

        Returns:
            pd.DataFrame: The result of the provided query.
        """
        return pd.read_sql(*self.args, sql=self.sql, con=self.con, **self.kwargs)



from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.0"

    deps = [
        f"mlpype-base=={version}",
        "numpy>=1.23.0",
        "scikit-learn>=1.1.1",
        "pandas>=1.4.3",
    ]
    strict_deps = [s.replace(">=", "==") for s in deps]

    setup(
        name="mlpype-sklearn",
        install_requires=deps,
        extras_require={"dev": strict_deps, "strict": strict_deps},
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )



from typing import Any, Dict

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession

from mlpype.base.data.data_source import DataSource


class SparkReadSource(DataSource[SparkDataFrame]):
    def __init__(
        self,
        spark_session: SparkSession,
        file: str,
        format: str,
        options: Dict[str, Any],
    ) -> None:
        """Read a file through Spark.

        Args:
            spark_session (SparkSession): The active SparkSessionl.
            file (str): The file path.
            format (str): The file format.
            options (Dict[str, Any]): Any additional options for `load`.
        """
        super().__init__()
        self.spark_session = spark_session
        self.file = file
        self.format = format
        self.options = options

    def read(self) -> SparkDataFrame:
        """Read the file at the given location.

        Returns:
            SparkDataFrame: A DataFrame based on the given file.
        """
        return self.spark_session.read.format(self.format).load(self.file, **self.options)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Union


class Serialiser(ABC):
    def __init__(self) -> None:
        """A base class for Serialisers, used to log extra files in Experiments.

        Please make sure any Serialiser you create can be deserialised using a JoblibSerialiser.
        These are used in the Experiment/Inferencer to save/load your Serialisers.
        """
        super().__init__()

    @abstractmethod
    def serialise(self, object: Any, file: Union[str, Path]) -> None:
        """Serialise the given object to the given file.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The file to serialise to.
        """
        raise NotImplementedError

    @abstractmethod
    def deserialise(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        Args:
            file (Union[str, Path]): The file to deserialise.

        Returns:
            Any: The python object stored in the file.
        """
        raise NotImplementedError



from typing import Any, Dict, List, Optional, Type, Union

from pyspark.ml import Estimator, Transformer

from mlpype.base.data.dataset import DataSet
from mlpype.base.pipeline.operator import Operator
from mlpype.base.pipeline.pipe import Pipe


class SparkPipe(Pipe):
    def __init__(
        self,
        name: str,
        operator: Type[Operator],
        inputs: List[str],
        outputs: List[str],
        kw_args: Optional[Dict[str, Any]] = None,
        fit_inputs: Optional[List[str]] = None,
        skip_on_inference: bool = False,
    ) -> None:
        """Same init as Pipe."""
        super().__init__(name, operator, inputs, outputs, kw_args, fit_inputs, skip_on_inference)
        self.fitted = None

    def fit(self, data: DataSet) -> "SparkPipe":
        """Fits the SparkPipe to the given DataSet.

        Args:
            data (DataSet): The DataSet to use in fitting.

        Returns:
            SparkPipe: This object.
        """
        op: Union[Transformer, Estimator] = self.operator

        if isinstance(op, Transformer):
            self.fitted = op
        elif isinstance(op, Estimator):
            self.fitted = op.fit(*data.get_all(self.inputs), *data.get_all(self.fit_inputs))
        else:
            raise ValueError(f"In a SparkPipe, the operator must be a Transformer or Estimator. Got: {type(op)}")
        return self

    def transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Transforms the given data using this Pipe.

        This Pipe should be fitted first using fit().
        This is version of Pipe is changed to work with Spark's API: transformation is done using the
        fitted object instead of the Operator directly.

        Args:
            data (DataSet): The DataSet to use in transforming.
            is_inference (Optional[bool]): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The transformed Data.
        """
        assert self.fitted is not None, "Make sure you fit the pipeline before transforming."

        # skip this step if we're in inference mode and this Pipe is marked as such.
        if is_inference and self.skip_on_inference:
            return data

        transformed = self.fitted.transform(*data.get_all(self.inputs))
        if len(self.outputs) < 2:
            transformed = [transformed]
        result = data.copy()
        result.set_all(self.outputs, transformed)
        return result

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Sets the state of this object from a dictionary.

        Used by pickle to properly prepare the "fitted" and "operator" fields as None's.

        Args:
            state (Dict[str, Any]): _description_
        """
        state["fitted"] = None
        state["operator"] = None
        self.__dict__ = state

    def __getstate__(self) -> Dict[str, Any]:
        """Gets the state of this object, excluding any Spark objects.

        This is to make sure serialisation works as expected.

        Returns:
            Dict[str, Any]: A dict representation of this object.
        """
        dct = self.__dict__.copy()
        if "fitted" in dct:
            del dct["fitted"]
        del dct["operator"]
        return dct

    def __str__(self) -> str:
        """Create string representation of this Pipe.

        Returns:
            str: A string representation of this Pipe.
        """
        fit_section = f" (+ {self.fit_inputs})" if len(self.fit_inputs) > 0 else ""
        input_output_section = f"{self.inputs}{fit_section} -> {self.outputs}"

        postfix = "unfitted" if self.fitted is None else str(self.fitted)

        return f"Pipe `{self.name}`, {input_output_section}: {postfix}"



import typing
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union

import numpy as np

from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_class
from mlpype.base.model import Model
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser
from mlpype.sklearn.data.sklearn_data import SklearnData
from mlpype.sklearn.model.sklearn_base_type import SklearnModelBaseType

T = TypeVar("T", bound=SklearnModelBaseType)


class SklearnModel(Model[SklearnData], ABC, Generic[T]):
    SKLEARN_MODEL_FILE = "model.pkl"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        model: Optional[T] = None,
        seed: int = 1,
        **model_args: Any,
    ) -> None:
        """A generic class for sklearn-like Models.

        You can set a sklearn model as a type hint to this class when defining a new model.
        This allows us to get the parameters from the documentation of that sklearn model.
        For an example, see the implementation of LinearModel, especially the `SklearnModel[LinearRegression]` part.

        class LinearRegressionModel(SklearnModel[LinearRegression]):
            def _init_model(self, **args) -> LinearRegression:
                return LinearRegression(**args)


        Args:
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            model (Optional[SklearnModelBaseType]): An object that has fit() and predict() methods. If none,
                we will use the model_args to instantiate a new model.
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
            **model_args (Any): Optional keyword arguments passed to the model class to instantiate a new
                model if `model` is None.
        """
        super().__init__(inputs, outputs, seed)
        if model is None:
            model = self._init_model(model_args)
        self.model = model

    @abstractmethod
    def _init_model(self, args: Dict[str, Any]) -> T:
        raise NotImplementedError

    @classmethod
    def _get_annotated_class(cls) -> Type[SklearnModelBaseType]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def set_seed(self) -> None:
        """Sets the RNG seed."""
        np.random.seed(self.seed)

    def _fit(self, *data: SklearnData) -> None:
        self.model.fit(*data)

    def _transform(self, *data: SklearnData) -> Union[Iterable[SklearnData], SklearnData]:
        return self.model.predict(*data)

    def _save(self, folder: Path) -> None:
        serialiser = JoblibSerialiser()
        serialiser.serialise(self.model, folder / self.SKLEARN_MODEL_FILE)

    @classmethod
    def _load(cls: Type["SklearnModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SklearnModel":
        serialiser = JoblibSerialiser()
        model = serialiser.deserialise(folder / cls.SKLEARN_MODEL_FILE)
        return cls(inputs=inputs, outputs=outputs, model=model, seed=1)

    @classmethod
    def get_parameters(cls: Type["SklearnModel"], parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        SklearnModel's will work by requiring 2 ways to instantiate a Model:
            - through `model`, which is a sklearn model.
            - through parameters, which will instantiate the model internally.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """
        super().get_parameters(parser)
        BaseModel = cls._get_annotated_class()

        add_args_to_parser_for_class(
            parser, BaseModel, "model", [], excluded_args=["seed", "inputs", "outputs", "model"]
        )



import subprocess
import sys
from pathlib import Path
from shutil import copyfile, copytree
from tempfile import TemporaryDirectory
from typing import List, Optional, Union

from mlpype.base.constants import Constants
from mlpype.base.utils.workspace import switch_directory


def make_wheel(
    model_folder: Path,
    model_name: str,
    version: str,
    output_wheel_file: Optional[Union[Path, str]] = None,
    libraries: Optional[List[str]] = None,
) -> None:
    """Turns an output folder from an Experiment into a wheel file.

    This should allow you to save models into shippable packages
    quite easily.

    Args:
        model_folder (Path): Path to the output folder generated by an Experiment
        model_name (str): The name of the model. Will be used as package name. Should follow
            proper package naming conventions.
        version (str): Model/package version.
        output_wheel_file (Optional[Union[Path, str]]): The output directory. Defaults to "wheel_output"
            in the current directory.
        libraries (Optional[List[str]]): A list of (versioned) libraries that are used as
            dependencies. By default we use the requirements.txt file from the experiment run.
    """
    if output_wheel_file is None:
        output_wheel_file = Path("").parent.absolute() / "wheel_output"

    local_dir = Path(__file__).parent / "helpers"
    template_file = local_dir / "setup_template.py"
    main_file = local_dir / "wheel_main.py"

    if libraries is None:
        with open(model_folder / Constants.REQUIREMENTS_FILE, "r") as f:
            requirements_raw = f.readlines()
            assert len(requirements_raw) > 0, "Using mlpype, it is assumend there is at least 1 dependency"
            requirements = '", "'.join(filter(lambda x: "==" in x, requirements_raw)).replace("\n", "")
    else:
        assert len(libraries) > 0, "Using mlpype, it is assumend there is at least 1 dependency"
        requirements = '", "'.join(libraries)

    with open(template_file, "r") as f:
        setup_file_raw = f.read()
        setup_file_formatted = (
            setup_file_raw.replace("{install_requires}", requirements)
            .replace("{package_name}", model_name)
            .replace("{version}", version)
        )

    with TemporaryDirectory() as tmp_directory:
        tmp_directory_path = Path(tmp_directory)
        tmp_model_dir = tmp_directory_path / model_name / "outputs"
        tmp_setup_file = tmp_directory_path / "setup.py"
        tmp_main_file = tmp_directory_path / model_name / "main.py"

        copytree(model_folder, tmp_model_dir)
        copyfile(main_file, tmp_main_file)

        with open(tmp_setup_file, "w") as f:
            f.write(setup_file_formatted)
        with open(tmp_directory_path / "__init__.py", "w") as f:
            f.write("")
        with open(tmp_directory_path / model_name / "__init__.py", "w") as f:
            f.write("from .main import load_model, load_app")

        output_wheel_file = str(output_wheel_file)
        with switch_directory(tmp_directory):
            subprocess.check_output(
                [sys.executable, "setup.py", "bdist_wheel", "--dist-dir", output_wheel_file]
            ).decode()



from pyspark.sql import DataFrame as SparkDataFrame

from mlpype.base.data.data_source import DataSource


class SparkDataFrameSource(DataSource[SparkDataFrame]):
    # should only be used for testing purposes!
    def __init__(
        self,
        df: SparkDataFrame,
    ) -> None:
        """Returns a fixed spark dataframe reference.

        This should only be used for testing/debugging purposes.
        Consider the SQL or Read classes for proper data reading!

        Args:
            df (SparkDataFrame): The DataFrame to use as output when
                read() is called.
        """
        super().__init__()
        self.df = df

    def read(self) -> SparkDataFrame:
        """Returns the stored DataFrame.

        Returns:
            SparkDataFrame: The stored DataFrame.
        """
        return self.df



from typing import Any, Dict, List, Optional, Union

from mlpype.base.data import DataSet
from mlpype.base.pipeline.pipe import Pipe
from mlpype.base.utils.parsing import get_args_for_prefix


class Pipeline:
    def __init__(self, pipes: List[Union[Pipe, "Pipeline"]]) -> None:
        """A pipeline of operations that can be re-applied to new, similar Data.

        Args:
            pipes (List[Union[Pipe, Pipeline]]): A list of either Pipes
                or other Pipelines. These form the steps that should be applied.
        """
        self.pipes = pipes
        self._assert_all_names_different()

    def _assert_all_names_different(self, names: Optional[set] = None) -> None:
        if names is None:
            names = set()

        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                pipe._assert_all_names_different(names)
            else:
                assert pipe.name not in names, f"{pipe.name} is used multiple times."
                names.add(pipe.name)

    def fit(self, data: DataSet) -> "Pipeline":
        """Fits the entire Pipeline to the given DataSet.

        Args:
            data (DataSet): The DataSet to use in fitting.

        Returns:
            Pipeline: This object.
        """
        for pipe in self.pipes:
            pipe.fit(data)
            data = pipe.transform(data, is_inference=False)
        return self

    def transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Transforms the given data using this Pipeline.

        This Pipeline should be fitted first using fit().

        Args:
            data (DataSet): The DataSet to use in transforming.
            is_inference (Optional[bool]): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The Transformed Data.
        """
        for pipe in self.pipes:
            data = pipe.transform(data, is_inference=is_inference)
        return data

    def inverse_transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Inverse transforms the DataSet using this Pipeline.

        Note that this is automatically done in reverse: the inverse steps
        will be applied from back to front.

        Args:
            data (DataSet): The DataSet to use in inverse transforming.
            is_inference (Optional[bool]): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The inverse transformed DataSet
        """
        for pipe in reversed(self.pipes):
            data = pipe.inverse_transform(data, is_inference=is_inference)
        return data

    def reinitialise(self, args: Dict[str, Any]) -> None:
        """Re-initialises this Pipeline's Pipes using the given dictionary.

        Args:
            args (Dict[str, Any]): Dictionary containing new arguments.
                The keys of the `args` dict should be of the form `<name>__<arg_name>`:
                - name: the name of the Pipe.
                - arg_name: the argument name to use.
                The values will be used as argument parameters.
        """
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                pipe.reinitialise(args)
            else:
                pipe_args = get_args_for_prefix(f"{pipe.name}__", args)
                if len(pipe_args) > 0:
                    pipe.reinitialise(pipe_args)

    def copy(self, args: Optional[Dict[str, Any]] = None) -> "Pipeline":
        """Create a copy of this Pipeline using the given dictionary.

        Args:
            args (Dict[str, Any]): Dictionary containing new arguments.
                The keys of the `args` dict should be of the form `<name>__<arg_name>`:
                - name: the name of the Pipe.
                - arg_name: the argument name to use.
                The values will be used as argument parameters.
        """
        result = Pipeline([])
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                extra_pipeline = pipe.copy(args)
                result += extra_pipeline
            else:
                if args is None:
                    pipe_args = {}
                else:
                    pipe_args = get_args_for_prefix(f"{pipe.name}__", args)
                new_pipe = pipe.copy(pipe_args)
                result += new_pipe
        return result

    def __iter__(self) -> "Pipeline":
        """Prepares this object for iteration using next().

        Returns:
            Pipeline: this object.
        """
        self.__iter_n = 0
        return self

    def __next__(self) -> Union["Pipeline", Pipe]:
        """Gets the next item in this Pipeline.

        Raises:
            StopIteration: If this object has already been iterated.

        Returns:
            Union[Pipeline, Pipe]: The next Pipeline/Pipe.
        """
        if self.__iter_n >= len(self.pipes):
            raise StopIteration
        else:
            pos = self.__iter_n
            self.__iter_n += 1
            return self[pos]

    def __getitem__(self, pos: Union[int, str]) -> Union["Pipeline", Pipe]:
        """Gets the Pipe/Pipeline at the given position.

        Raises:
            KeyError: if `pos` does not exist in this object.
                int: `pos` is out of bounds.
                str: the given name does not match any Pipe in this object.
                    Any Pipelines in this object are also searched. This makes it
                    important to use different names for Pipes!

        Returns:
            Union[Pipeline, Pipe]: The Pipeline/Pipe matching the given position.
        """
        value = self._get(pos)
        if value is None:
            raise KeyError(f"{pos} is not found in this pipeline")
        return value

    def _get(self, pos: Union[int, str]) -> Union["Pipeline", Pipe, None]:
        if isinstance(pos, int):
            return self.pipes[pos]
        for pipe in self.pipes:
            if isinstance(pipe, Pipeline):
                result = pipe._get(pos)
                if result is not None:
                    return result
            elif pipe.name == pos:
                return pipe
        return None

    def __add__(self, obj: Union["Pipeline", Pipe]) -> "Pipeline":
        """Append a Pipe or Pipeline object and return a new Pipeline.

        Args:
            obj (Union[Pipeline, Pipe]): The object to be appended. Can be a Union[Pipeline, Pipe].

        Raises:
            TypeError: If `obj` is not a Pipe or Pipeline.

        Returns:
            Pipeline: A Pipeline with the two sets of steps combined. The steps from this
                object will always be used first.
        """
        if isinstance(obj, Pipeline):
            return Pipeline(self.pipes + obj.pipes)
        elif isinstance(obj, Pipe):
            return Pipeline(self.pipes + [obj])
        else:
            raise TypeError("Pipeline can only use + with a Pipe or Pipeline")

    def insert(self, pos: int, pipe: Pipe) -> None:
        """Insert a pipe in the given integer position.

        We check if the pipe name doesn't exist yet.

        Args:
            pos (int): The position to put the new pipe
            pipe (Pipe): The pipe to insert
        """
        self._assert_all_names_different(set(pipe.name))
        self.pipes.insert(pos, pipe)

    def __len__(self) -> int:
        """Return the length of this Pipeline.

        If it contains any sub-pipelines, those will be counted in full as well
        (and thus not as a single step, unless they only contain 1 step).

        Returns:
            int: The lenght of this Pipeline.
        """
        result = 0
        for p in self.pipes:
            if isinstance(p, Pipeline):
                result += len(p)
            else:
                result += 1
        return result

    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Pipeline.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Pipeline.
        """
        tabs = "\t" * indents
        output = [tabs + "Pipeline:"]
        indents += 1
        for pipe in self.pipes:
            if isinstance(pipe, Pipe):
                output.append(tabs + str(pipe))
            else:
                output.append(pipe.__str__(indents))
        return "\n".join(output)



import json
from abc import ABC, abstractclassmethod, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import Generic, List, Tuple, TypeVar, Union

from mlpype.base.constants import Constants
from mlpype.base.data import DataSet
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser

Data = TypeVar("Data")


class Model(ABC, Generic[Data]):
    def __init__(self, inputs: List[str], outputs: List[str], seed: int = 1) -> None:
        """An abstraction of a ML model.

        This should provide the basic interface for fitting, inference, and serialisation.

        Args:
            inputs (List[str]): A list of names of input Data. This determines which Data is
                used to fit the model.
            outputs (List[str]): A list of names of output Data. This determines the names of
                output variables.
            seed (int, optional): The RNG seed to ensure reproducability.. Defaults to 1.
        """
        super().__init__()
        self.seed = seed
        self.inputs = inputs
        self.outputs = outputs
        self.set_seed()

    @classmethod
    def get_parameters(cls, parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """

    @abstractmethod
    def set_seed(self) -> None:
        """Sets the RNG seed."""
        raise NotImplementedError

    def save(self, folder: Union[str, Path]) -> None:
        """Stores this model to the given folder.

        This function stores the common inputs and outputs list to the given folder,
        and makes sure it's created. It also stores the model class in the given folder.
        It will call _save to allow further models to specify how they are stored.

        Args:
            folder (Union[str, Path]): The folder to store the Model in.
        """
        joblib_serialiser = JoblibSerialiser()

        if isinstance(folder, str):
            folder = Path(folder)

        folder.mkdir(exist_ok=True, parents=True)
        lists_to_save = {
            "inputs": self.inputs,
            "outputs": self.outputs,
        }
        with open(folder / Constants.MODEL_PARAM_FILE, "w") as f:
            json.dump(lists_to_save, f)

        joblib_serialiser.serialise(self.__class__, folder / Constants.MODEL_CLASS_FILE)
        self._save(folder)

    @abstractmethod
    def _save(self, folder: Path) -> None:
        """Stores this model to the given folder.

        Specifically intended to store

        Args:
            folder (Path): The folder to store the Model in.
        """
        raise NotImplementedError

    @classmethod
    def load(cls, folder: Union[str, Path]) -> "Model":
        """Loads a model from file into this Model.

        Args:
            folder (Union[str, Path]): The folder to load the model from.
        """
        if isinstance(folder, str):
            folder = Path(folder)

        with open(folder / Constants.MODEL_PARAM_FILE) as f:
            lists = json.load(f)

        joblib_serialiser = JoblibSerialiser()
        model_class = joblib_serialiser.deserialise(folder / Constants.MODEL_CLASS_FILE)
        return model_class._load(folder, lists["inputs"], lists["outputs"])

    @abstractclassmethod
    def _load(cls, folder: Path, inputs: List[str], outputs: List[str]) -> "Model":
        """Loads a model from file into this Model.

        Args:
            folder (Union[str, Path]): The folder to load the model from.
        """
        raise NotImplementedError

    def fit(self, data: DataSet) -> "Model":
        """Fits the Model to the given DataSet.

        The DataSet should contain all inputs and outputs.

        Args:
            data (DataSet): The DataSet to fit this Model on.
        """
        self._fit(*data.get_all(self.inputs), *data.get_all(self.outputs))
        return self

    @abstractmethod
    def _fit(self, *data: Data) -> None:
        raise NotImplementedError

    def transform(self, data: DataSet) -> DataSet:
        """Applies the Model to the given DataSet.

        The DataSet should contain all inputs.

        Args:
            data (DataSet): The DataSet to transform using this Model.

        Returns:
            DataSet: The outputs as a Dataset.
        """
        result = self._transform(*data.get_all(self.inputs))
        if len(self.outputs) == 1:
            result = [result]  # type: ignore

        # type check handled by above check.
        return DataSet.from_dict({name: data for name, data in zip(self.outputs, result)})  # type: ignore

    @abstractmethod
    def _transform(self, *data: Data) -> Union[Tuple[Data], Data]:
        raise NotImplementedError

    def __str__(self) -> str:
        """Create string representation of this Model.

        Returns:
            str: A string representation of this Model.
        """
        return f"{type(self).__name__}: {self.inputs} -> {self.outputs}"



from typing import List

from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.0"
    deps: List[str] = [f"mlpype-base=={version}", "hyperopt>=0.2.7"]
    strict_deps = [s.replace(">=", "==") for s in deps]

    setup(
        name="mlpype-hyperopt",
        install_requires=deps,
        extras_require={"dev": strict_deps, "strict": strict_deps},
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )



from abc import ABC, abstractmethod
from typing import Any, Iterable, Union

from mlpype.sklearn.data.sklearn_data import SklearnData


class SklearnModelBaseType(ABC):
    @abstractmethod
    def fit(self, *x: SklearnData) -> Any:
        """Fit a model to the given data."""

    @abstractmethod
    def predict(self, *x: SklearnData) -> Union[Iterable[SklearnData], SklearnData]:
        """Predict for given data using a trained model."""



class Constants:
    MODEL_FOLDER: str = "model"
    MODEL_CLASS_FILE: str = "model_class"
    MODEL_PARAM_FILE: str = "model_inputs_outputs.json"
    PIPELINE_FILE: str = "pipeline.pkl"
    INPUT_TYPE_CHECKER_FILE: str = "input_type_checker.pkl"
    OUTPUT_TYPE_CHECKER_FILE: str = "output_type_checker.pkl"
    EXTRA_FILES: str = "extra_files.json"
    SERIALISER_FILE: str = "serialiser.pkl"

    PYTHON_VERSION_FILE: str = "python_version.json"
    REQUIREMENTS_FILE: str = "requirements.txt"



import inspect
import re
import typing
import warnings
from argparse import ArgumentParser
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union

from docstring_parser import parse

from mlpype.base.pipeline.pipe import Pipe
from mlpype.base.pipeline.pipeline import Pipeline


def add_args_to_parser_for_pipeline(parser: ArgumentParser, pipeline: Pipeline) -> None:
    """Adds arguments for a full pipeline re-initialisation to the given ArgumentParser.

    This will add arguments for every Pipe in the pipeline.

    Arguments will be drawn in order from:
        - type hints
        - documentation.

    Documentation hints are not guaranteed to work. For the most reliable results, use type hints.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        pipeline (Pipeline): The pipeline for which re-initialisation arguments should be added.
    """
    for pipe in pipeline:
        if isinstance(pipe, Pipe):
            add_args_to_parser_for_class(parser, pipe.operator_class, f"pipeline__{pipe.name}", [])
        else:
            add_args_to_parser_for_pipeline(parser, pipe)


def add_args_to_parser_for_class(
    parser: ArgumentParser,
    class_: Type,
    prefix: str,
    excluded_superclasses: List[Type],
    excluded_args: Optional[List[str]] = None,
) -> None:
    """Adds arguments for the constructor of a class to the given parser.

    We try to dive deeper in case of variable kw-args, but in general we recommend not
    using these.

    Arguments will be drawn in order from:
        - type hints
        - documentation.

    Documentation hints are not guaranteed to work. For the most reliable results, use type hints.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        class_ (Type): The class who's constructor's arguments we want to add.
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        excluded_superclasses (List[Type]): Superclasses of the class_ that should be ignored in case
            there are kw-args style arguments to the constructor.
        excluded_args (Optional[List[str]], optional): argument names to never include.
            By default 'self' and 'cls' are ignored.
    """
    init_func = class_.__init__
    class_docstring_args = _parse_docs_to_type_args(class_)
    if len(class_docstring_args) == 0:
        class_docstring_args = _parse_docs_to_type_args(init_func)

    add_args_to_parser_for_function(parser, init_func, prefix, excluded_args, class_docstring_args=class_docstring_args)

    signature = inspect.signature(init_func)
    for _, param in signature.parameters.items():
        # we only need to check for superclass arguments if there are var-args.
        # we ignore positional var-args.
        if param.kind == param.VAR_KEYWORD:
            for superclass in class_.__bases__:
                if superclass not in excluded_superclasses:
                    add_args_to_parser_for_class(parser, superclass, prefix, excluded_superclasses, excluded_args)


def add_args_to_parser_for_function(
    parser: ArgumentParser,
    function: Callable,
    prefix: str,
    excluded: Optional[List[str]] = None,
    class_docstring_args: Optional[Dict[str, Union[type, None]]] = None,
) -> None:
    """Add arguments for a given function to the given parser.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        function (Callable): The function who's arguments we want to add.
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        excluded (Optional[List[str]]): argument names to never include. By default 'self' and 'cls' are ignored.
        class_docstring_args: (Optional[Dict[str, Union[type, None]]): docstring arguments obtained from the class.
            Useful for init functions, since those can be documented in multiple places.
    """
    args = inspect.signature(function)
    docstring_args = _parse_docs_to_type_args(function)

    if class_docstring_args is not None:
        docstring_args.update(**class_docstring_args)

    for name, parameter in args.parameters.items():
        class_ = parameter.annotation
        if class_ == inspect._empty and name in docstring_args:
            class_ = docstring_args[name]
        required = parameter.default == inspect._empty
        add_argument(parser, name, prefix, class_, required, parameter.default, excluded)


def _parse_docs_to_type_args(
    func: Callable, extra_mappings: Optional[Dict[str, type]] = None, include_none_args: bool = False
) -> Dict[str, Union[type, None]]:
    """Parses the docstring of a function into an arg-type dictionary.

    We only parse the following types, whose first letter can be capitalized.
        - bool
        - str
        - int
        - float

    This is all done on a best-effort basis. We do not guarantee this will work for every
    function in every library. For the best results, make sure you use proper type hinting
    instead.

    Args:
        func (Callable): The function whose docstring needs to be read.
        extra_mappings (Dict[str, type], optional): Optional extra regex -> type mappings.
            Defaults to no extra mappings.
        include_none_args (bool, optional): Whether to include arguments in the output
            with no good type mappings. Defaults to False.

    Returns:
        Dict[str, Union[type, None]]: A dictionary mapping argument names to types.
    """
    signature = inspect.signature(func)
    if extra_mappings is None:
        extra_mappings = {}

    result: Dict[str, Union[type, None]] = {}
    doc = inspect.getdoc(func)
    if doc is None:
        return result

    parsed_doc = parse(doc)
    for param in parsed_doc.params:
        if param.arg_name in signature.parameters:
            class_ = _parse_type_name(param.type_name, extra_mappings)
            if include_none_args or class_ is not None:
                result[param.arg_name] = class_

    return result


def _parse_type_name(s: Optional[str], extra_mappings: Optional[Dict[str, type]] = None) -> Union[type, None]:
    if extra_mappings is None:
        extra_mappings = {}
    if s is None:
        return None

    regexes = {
        r"(^|\W)[fF]loat($|\W)": float,
        r"(^|\W)[sS]tr($|\W)": str,
        r"(^|\W)[bB]ool($|\W)": bool,
        r"(^|\W)[iI]nt($|\W)": int,
    }
    regexes.update(**extra_mappings)

    for regex, class_ in regexes.items():
        if re.search(regex, s) is not None:
            return class_
    return None


def add_argument(
    parser: ArgumentParser,
    name: str,
    prefix: str,
    class_: Union[type, None],
    is_required: bool,
    default: Any,
    excluded: Optional[List[str]] = None,
) -> None:
    """Add an argument to the given parser.

    Args:
        parser (ArgumentParser): The ArgumentParser to add arguments to.
        name (str): The name of the argument
        prefix (str): A prefix that should be set on each argument name before adding it to
            the ArgumentParser.
        class_ (type): The class of the argument. Currently we support:
            - str, float, int, bool
            - lists / tuples / iterables of the above 4 arguments.
        is_required (bool): Whether the argument is required or optional.
        default (Any): the default value if is_required is False.
        excluded (Optional[List[str]]): argument names to never include. By default 'self' and 'cls' are ignored.
    """
    if excluded is None:
        excluded = []
    excluded = excluded + ["self", "cls"]

    if not is_required:
        assert default != inspect._empty, "A default value must be provided if the argument is not required"
    else:
        default = None

    arg_name = f"--{prefix}__{name}"
    if name in excluded or class_ is None:
        return
    if class_ in [str, float, int, bool]:
        parser.add_argument(arg_name, type=_get_conversion_function(class_), required=is_required, default=default)
    elif typing.get_origin(class_) in [
        list,
        tuple,
        typing.get_origin(Iterable),
        typing.get_origin(List),
        typing.get_origin(Tuple),
    ]:
        subtype = typing.get_args(class_)[0]
        parser.add_argument(
            arg_name, type=_get_conversion_function(subtype), nargs="+", required=is_required, default=default
        )
    else:
        warnings.warn(f"Currently the class {str(class_)} is not supported for automatic command line importing.")


def _get_conversion_function(class_: type) -> Callable:
    if class_ == bool:
        return _convert_bool
    return class_


def _convert_bool(s: str) -> bool:
    return s.lower() in ["true", "1"]



from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from .experiment_logger import ExperimentLogger


class LocalLogger(ExperimentLogger):
    def __enter__(self) -> None:
        """Start the experiment."""

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """Stop the experiment."""

    def __init__(self, min_whitespace: int = 30) -> None:
        """A basic logger which only saves local copies of the data and prints results."""
        super().__init__()
        self.min_whitespace = min_whitespace

    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Prints the metrics to the console.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        for name, value in metrics.items():
            print(f"{name.ljust(self.min_whitespace)}: {value}")

    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Prints the parameters to the console.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        for name, value in parameters.items():
            print(f"{name.ljust(self.min_whitespace)}: {value}")

    def log_file(self, file: Union[str, Path]) -> None:
        """This method does no additional logging.

        Args:
            file (Union[str, Path]): The file to log.
        """



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union

from mlflow.artifacts import download_artifacts
from mlflow.tracking import set_tracking_uri
from mlflow.tracking.fluent import get_experiment_by_name, list_run_infos

from mlpype.base.deploy import Inferencer


def load_experiment_from_mlflow(
    url: str,
    experiment_name: str,
    run_id: str,
    directory: Optional[Union[Path, str]] = None,
) -> Inferencer:
    """Download and import a trained mlpype Model from mlflow's artifact store.

    Args:
        url (str): The tracking url.
        experiment_name (str): The name of the experiment. Used to verify that the
            run_id is the correct one.
        run_id (str): The run id. Used to pull the actual experiment.
        directory (Optional[Union[Path, str]]): Optional directory to store the results
            in. By default, a temporary directory is used.

    Returns:
        Inferencer: _description_
    """
    set_tracking_uri(url)

    # verify the run ID provided is part of the given experiment.
    exp_id = get_experiment_by_name(experiment_name)
    assert exp_id is not None, f"Experiment {experiment_name} does not exist in {url}."
    assert run_id in [
        run.run_id for run in list_run_infos(exp_id.experiment_id)
    ], f"Run ID {run_id} is not present in the given experiment."

    if directory is None:
        with TemporaryDirectory(prefix="mlflow_model") as tmp_dir:
            return _download_and_load(run_id, Path(tmp_dir))
    else:
        if isinstance(directory, str):
            directory = Path(directory)
        return _download_and_load(run_id, directory)


def _download_and_load(
    run_id: str,
    directory: Path,
) -> Inferencer:
    download_artifacts(f"runs:/{run_id}/", dst_path=str(directory))
    return Inferencer.from_folder(directory)



from typing import Callable, Dict, Optional, TypeVar, Union

from mlpype.base.data import DataSet
from mlpype.base.evaluate.base_evaluator import BaseEvaluator
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline

Data = TypeVar("Data")


class Evaluator(BaseEvaluator[Data]):
    def __init__(
        self,
        functions: Dict[str, Callable[[Union[Data, Data]], Union[float, int, str, bool]]],
    ) -> None:
        """Evaluates a Model on the given Functions.

        Args:
            functions (Dict[str, Callable[[Union[Data, Data]], Union[float, int, str, bool]]]):
                A dict of metric names and metric functions. We expect each to return
                a float, int, string, or boolean value. We provide the arguments as follows:
                    - first, all labeled data, e.g. y_true.
                    - second, all predicted data, e.g. y_pred.
        """
        self.functions = functions

    def evaluate(
        self, model: Model, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the model does not need to be transformed anymore if pipeline is None.

        Args:
            model (Model): The Model to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """
        if pipeline is not None:
            data = pipeline.transform(data)

        predictions = model.transform(data).get_all(model.outputs)
        output_data = data.get_all(model.outputs)

        return {name: func(*output_data, *predictions) for name, func in self.functions.items()}

    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Evaluator.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Evaluator.
        """
        tabs = "\t" * indents
        return "\n".join([f"{tabs}{name}: {func.__name__}" for name, func in self.functions.items()])



import re
from dataclasses import dataclass
from pathlib import Path
from shutil import copyfile
from typing import List, Tuple


@dataclass
class WheelExtension:
    """An extension to mlpypes deployment process in wheel files.

    This class provides standardised configuration to add plugins to mlpype wheel files.

    Extensions require 3 parameters:
        - name: A unique name for the extension. This will be used to split imports and files.
        - functionality: A list of (path, functions) pairs. These are files to be copied over to any
            wheel package being created. Please make sure any imports in these files are NOT relative
            and do not otherwise rely on file structure/ordering. Ideally you should make sure each
            file is self-contained, only relying on external dependencies.
        - libraries that should be installed as extra dependencies to make the plugin work.

    All extensions will be placed in a similar fashion inside the wheel file:
        - <name of the model/package>
        |
        --  - <plugin name 1>
                - files for plugin 1
            - <plugin name 2>
            - <plugin name 3>
    Files from a plugin will be place in a flat manner inside the folder. Duplicate file names will cause issues.
    """

    name: str
    functionality: List[Tuple[Path, List[str]]]
    libraries: List[str]

    def __post_init__(self) -> None:
        """Perform checks on inputs."""
        self._check_file_name(self.name)
        for path, _ in self.functionality:
            self._check_file_name(path.name.replace(".py", ""))

        files = set()
        function_names = set()

        for path, functions in self.functionality:
            assert path not in files, f"{path} is a duplicated file name, this is not allowed."
            files.add(path)

            for func in functions:
                assert (
                    func not in function_names
                ), f"{func} is a duplicated function name from {path}, this is not allowed."
                function_names.add(func)

    @staticmethod
    def _check_file_name(name: str) -> None:
        """Check that the given name is an okay filename for imports.

        Args:
            name (str): The name to check

        Raises:
            AssertionError: if the name starts with a non-letter or contains a space or dot.
        """
        assert (
            re.search(r"^[a-zA-Z]", name) is not None
        ), f"Wheel extension file names need to start with a letter: {name}"

        for character in [" ", "."]:
            assert character not in name, f"{character} is not supported in wheel extension names: {name}"

    def extend(
        self,
        package_name: str,
        libraries: List[str],
        wheel_package_dir: Path,
    ) -> None:
        """Extends the given wheel file with this extension.

        Args:
            package_name (str): The name of the main package.
            libraries (List[str]): The current set of libraries to be installed.
            wheel_package_dir (Path): The Path to the main package.
        """
        libraries.extend(self.libraries)

        init_file_path = wheel_package_dir / "__init__.py"
        subfolder = wheel_package_dir / self.name
        subfolder.mkdir(exist_ok=False)

        with open(subfolder / "__init__.py", "w") as f:
            f.write("")

        with open(init_file_path, "a") as init_file:
            for file, imports in self.functionality:
                copyfile(file, subfolder / file.name)
                import_file_name = file.name.replace(".py", "")
                imports_str = ", ".join(imports)

                init_line = f"from {package_name}.{self.name}.{import_file_name} import {imports_str}\n"
                init_file.write(init_line)



from datetime import datetime
from typing import Any, Callable, Dict, List, Tuple, Type, Union

import numpy as np
import pandas as pd
from pandas.api.types import (
    is_bool_dtype,
    is_datetime64_any_dtype,
    is_float_dtype,
    is_integer_dtype,
    is_object_dtype,
    is_string_dtype,
)
from pydantic import create_model

from mlpype.base.pipeline.type_checker import DataModel, TypeChecker


class PandasData(DataModel):
    def convert(self) -> pd.DataFrame:
        """Converts this object to a pandas DataFrame.

        Returns:
            pd.DataFrame: The pandas DataFrame contained by this object.
        """
        return pd.DataFrame(data=self.__dict__)

    @classmethod
    def to_model(cls, data: pd.DataFrame) -> "PandasData":
        """Converts a pandas DataFrame to a PandasData model, which can be serialised.

        Args:
            data (pd.DataFrame): A pandas DataFrame to serialise.

        Returns:
            PandasData: A serialisable version of the DataFrame.
        """
        return cls(**{name: [str(i) for i in data[name].to_list()] for name in data.columns})


class PandasTypeChecker(TypeChecker[pd.DataFrame]):
    def fit(self, data: pd.DataFrame) -> "PandasTypeChecker":
        """Fit this PandasTypeChecker to the given data.

        Args:
            data (pd.DataFrame): The data to fit.

        Returns:
            PandasTypeChecker: self
        """
        self.raw_types = self._convert_raw_types(dict(data.dtypes))
        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Checks if the given data fits the specifications this TypeChecker was fitted for.

        Args:
            data (pd.DataFrame): The data to check.

        Returns:
            pd.DataFrame: data, if the data fits the specifications. Otherwise, an assertion error is thrown.
        """
        assert isinstance(data, pd.DataFrame), "Please provide a pandas DataFrame!"
        colnames = list(self.raw_types.keys())
        assert np.all(np.isin(colnames, data.columns)), "Not all columns are present."

        data = data[colnames]

        for name, (_, checker) in self.raw_types.items():
            assert checker(data[name]), f"Dtypes did not match up for col {name}."
        return data

    def _convert_raw_types(self, types: Dict[str, type]) -> Dict[str, Tuple[type, Callable]]:
        return {name: self._convert_raw_type(type_) for name, type_ in types.items()}

    def _convert_raw_type(self, type_: type) -> Tuple[type, Callable]:
        str_type = str(type_)
        if "int" in str_type:
            return (int, is_integer_dtype)
        elif "float" in str_type:
            return (float, is_float_dtype)
        elif "bool" in str_type:
            return (bool, is_bool_dtype)
        elif "datetime" in str_type:
            return (datetime, is_datetime64_any_dtype)
        elif "str" in str_type:
            return (str, is_string_dtype)
        else:
            return (str, is_object_dtype)

    def get_pydantic_type(self) -> Type[PandasData]:
        """Creates a Pydantic model for this data to handle serialisation/deserialisation.

        Returns:
            Type[PandasData]: A PandasData model that fits the data this wat fitted on.
        """
        data_type = {
            name: (Union[List[dtype], Dict[str or int, dtype]], ...)  # type: ignore
            for name, (dtype, _) in self.raw_types.items()
        }

        model = create_model("PandasData", **data_type, __base__=PandasData)

        return model

    @classmethod
    def supports_object(cls, obj: Any) -> bool:
        """Returns True if the object is a pandas DataFrame.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: True if the given object is a pandas DataFrame, False otherwise.
        """
        return isinstance(obj, pd.DataFrame)



from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.0"

    dev_deps = ["pre-commit", "build==0.8.0", "pypiserver==1.5.1", "twine==4.0.1", "pdoc==13.1.0"]
    test_deps = ["pytest", "pytest-cov"]
    deps = [f"mlpype=={version}", "docstring_parser>=0.14.1", "pydantic>=1.9.1", "joblib>=1.1.0", "PyYAML==6.0"]
    strict_deps = [s.replace(">=", "==") for s in deps]

    setup(
        name="mlpype-base",
        install_requires=deps,
        extras_require={
            "dev": strict_deps + dev_deps + test_deps,
            "test": strict_deps + test_deps,
            "strict": strict_deps,
        },
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )



from .linear_regression_model import LinearRegressionModel
from .logistic_regression_model import LogisticRegressionModel
from .sklearn_base_type import SklearnModelBaseType
from .sklearn_model import SklearnModel



import pandas as pd

from mlpype.base.data.data_source import DataSource


class DataFrameSource(DataSource[pd.DataFrame]):
    def __init__(self, df: pd.DataFrame) -> None:
        """A DataSource based around a created DataFrame.

        Good for testing purposes, but please use other sources for actual
        ML runs to prevent any data having to be hard-coded into the script.

        Args:
            df (pd.DataFrame): The DataFrame to use as a source.
        """
        super().__init__()
        self.df = df

    def read(self) -> pd.DataFrame:
        """Returns the DataFrame given as input before.

        Returns:
            pd.DataFrame: The DataFrame used to initialise this object.
        """
        return self.df



from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.0"

    deps = [
        f"mlpype-base=={version}",
        # on mac, it is recommended to use conda/mamba or source to install tensorflow
        "tensorflow>=2.12",
        "numpy>=1.23.0",
        "protobuf>=3.20.3",
    ]
    strict_deps = [s.replace(">=", "==") for s in deps]

    setup(
        name="mlpype-tensorflow",
        install_requires=deps,
        extras_require={"dev": strict_deps, "strict": strict_deps},
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )
